package rocketBugsRender;
/*
	LinearFit ... the following class incorporates several "Numerical Recipes in C" 
	(Chapter 15) procedures for fitting a straight-line to data with errors.
*/

/*  <RocketBugsRender - for rendering and graphing the output of Zoom (and other programs)>
    Copyright (C) <2007>  <Jonathan B. Alberts>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.Math.*;

public class LinearFit {

	
	public LinearFit () {
	}
	
	public static Line fit1D (double [] x, double [] y, int ndata, double [] sig, boolean mwt) {
		/*
		Given a set of data points x[1..ndata],y[1..ndata] with individual standard deviations
		sig[1..ndata], fit them to a straight line y = a + bx by minimizing q2. Returned are
		a,b and their respective probable uncertainties siga and sigb, the chi-square chiSq, and the
		goodness-of-fit probability q (that the fit would have q2 this large or larger). If mwt=true on
		input, then the standard deviations are assumed to be unavailable: q is returned as 1.0 and
		the normalization of chi2 is to unit standard deviation on all points.
		*/
		//talkln ("size of x is " + String.valueOf(x.length));
		//talkln ("ndata is " + String.valueOf(ndata));
		double wt,t,sxoss,sx=0.0,sy=0.0,st2=0.0,ss,sigdat;
		double a,b=0.0,sigA,sigB,chiSq,q;
		if (mwt) { //Accumulate sums ...
			ss=0.0;
			for (int i=0;i<ndata;i++) { //...with weights
				wt=1.0/Math.sqrt(sig[i]);
				ss += wt;
				sx += x[i]*wt;
				sy += y[i]*wt;
			}
		} else {
			for (int i=0;i<ndata;i++) { //...or without weights.
				sx += x[i];
				sy += y[i];
			}
			ss=(double)ndata;
		}
		sxoss=sx/ss;
		if (mwt) {
			for (int i=0;i<ndata;i++) {
				t=(x[i]-sxoss)/sig[i];
				st2 += t*t;
				b += t*y[i]/sig[i];
			}
		} else {
			for (int i=0;i<ndata;i++) {
				t=x[i]-sxoss;
				st2 += t*t;
				b += t*y[i];
			}
		}
		b /= st2; //Solve for a, b, sigA, and sigB.
		a=(sy-sx*b)/ss;
		sigA=Math.sqrt((1.0+sx*sx/(ss*st2))/ss);
		sigB=Math.sqrt(1.0/st2);
		chiSq=0.0; // Calculate chi squared.
		q=1.0;
		if (!mwt) {
			for (int i=0;i<ndata;i++) { chiSq += Math.pow((y[i]-a-b*x[i]),2); }
			sigdat = Math.sqrt(chiSq/((double)(ndata-2))); //For unweighted data evaluate typical sig using chi2, and adjust the standard deviations.
			sigA *= sigdat;
			sigB *= sigdat;
		} else {
			for (int i=0;i<ndata;i++) { chiSq += Math.pow(((y[i]-a-b*x[i])/sig[i]),2); }
			if (ndata>=2) q=NR.gammq(0.5*(ndata-2),0.5*chiSq); 	// Equation (15.2.12).
		}
		return new Line(b,a,(int)x[0],(int)x[ndata-1],sigB,sigA,chiSq,q);
	}
	
	private static void talk (String info) {
		System.out.print(info);
	}
	
	private static void talkln (String info) {
		System.out.println(info);
	}
	
}
	
/*	public double [] fitexy(double [] x, double [] y, int ndat, double [] sigx, double [] sigy) {
		double POTN = 1.57100;
		double BIG = 1.0e30;
		double PI = Math.PI;
		double ACC = 1.0e-3;
		double xx,yy,sx,sy,ww,aa,offs;	//Global
		int nn; //Global variables communicate with chixy. 
	}
	
	/*Straight-line fit to input data x[1..ndat] and y[1..ndat] with errors in both x and y, the respective
	standard deviations being the input quantities sigx[1..ndat] and sigy[1..ndat].
	Output quantities are a and b such that y = a + bx minimizes q2, whose value is returned
	as chi2. The q2 probability is returned as q, a small value indicating a poor fit (sometimes
	indicating underestimated errors). Standard errors on a and b are returned as siga and sigb.
	These are not meaningful if either (i) the fit is poor, or (ii) b is so large that the data are
	consistent with a vertical (infinite b) line. If siga and sigb are returned as BIG, then the data
	are consistent with all values of b.
	*/
/*void avevar(float data[], unsigned long n, float *ave, float *var);
float brent(float ax, float bx, float cx,
float (*f)(float), float tol, float *xmin);
float chixy(float bang);
void fit(float x[], float y[], int ndata, float sig[], int mwt,
float *a, float *b, float *siga, float *sigb, float *chi2, float *q);
float gammq(float a, float x);
void mnbrak(float *ax, float *bx, float *cx, float *fa, float *fb,
float *fc, float (*func)(float));
float zbrent(float (*func)(float), float x1, float x2, float tol);
int j;
float swap,amx,amn,varx,vary,ang[7],ch[7],scale,bmn,bmx,d1,d2,r2,
dum1,dum2,dum3,dum4,dum5;
xx=vector(1,ndat);
yy=vector(1,ndat);
sx=vector(1,ndat);
sy=vector(1,ndat);
ww=vector(1,ndat);
avevar(x,ndat,&dum1,&varx); //Find the x and y variances, and scale the data into the global variables for communication with the function chixy.
avevar(y,ndat,&dum1,&vary);
scale=sqrt(varx/vary);
nn=ndat;
for (j=1;j<=ndat;j++) {
xx[j]=x[j];
yy[j]=y[j]*scale;
sx[j]=sigx[j];
sy[j]=sigy[j]*scale;
ww[j]=sqrt(SQR(sx[j])+SQR(sy[j])); Use both x and y weights in first
trial fit. }
fit(xx,yy,nn,ww,1,&dum1,b,&dum2,&dum3,&dum4,&dum5); Trial fit for b.
offs=ang[1]=0.0; Construct several angles for reference
points, and make b an angle.
ang[2]=atan(*b);
ang[4]=0.0;
ang[5]=ang[2];
ang[6]=POTN;
for (j=4;j<=6;j++) ch[j]=chixy(ang[j]);
mnbrak(&ang[1],&ang[2],&ang[3],&ch[1],&ch[2],&ch[3],chixy);
Bracket the q2 minimum and then locate it with brent.
*chi2=brent(ang[1],ang[2],ang[3],chixy,ACC,b);
*chi2=chixy(*b);
*a=aa;
*q=gammq(0.5*(nn-2),*chi2*0.5); Compute q2 probability.
for (r2=0.0,j=1;j<=nn;j++) r2 += ww[j]; Save the inverse sum of weights at
the minimum. r2=1.0/r2;
bmx=BIG; Now, find standard errors for b as
points where ƣq2 = 1. bmn=BIG;
offs=(*chi2)+1.0;
for (j=1;j<=6;j++) { Go through saved values to bracket
the desired roots. Note periodicity
in slope angles.
if (ch[j] > offs) {
d1=fabs(ang[j]-(*b));
while (d1 >= PI) d1 -= PI;
d2=PI-d1;
if (ang[j] < *b) {
swap=d1;
d1=d2;
d2=swap;
}
if (d1 < bmx) bmx=d1;
if (d2 < bmn) bmn=d2;
}
}
if (bmx < BIG) { Call zbrent to find the roots.
bmx=zbrent(chixy,*b,*b+bmx,ACC)-(*b);
amx=aa-(*a);
bmn=zbrent(chixy,*b,*b-bmn,ACC)-(*b);
amn=aa-(*a);
*sigb=sqrt(0.5*(bmx*bmx+bmn*bmn))/(scale*SQR(cos(*b)));
*siga=sqrt(0.5*(amx*amx+amn*amn)+r2)/scale; Error in a has additional piece
r2. } else (*sigb)=(*siga)=BIG;
*a /= scale; Unscale the answers.
*b=tan(*b)/scale;
free_vector(ww,1,ndat);
free_vector(sy,1,ndat);
free_vector(sx,1,ndat);
free_vector(yy,1,ndat);
free_vector(xx,1,ndat);
}
670 Chapter 15. Modeling of Data
Sample page from NUMERICAL RECIPES IN C: THE ART OF SCIENTIFIC COMPUTING (ISBN 0-521-43108-5)
Copyright (C) 1988-1992 by Cambridge University Press. Programs Copyright (C) 1988-1992 by Numerical Recipes Software.
Permission is granted for internet users to make one paper copy for their own personal use. Further reproduction, or any copying of machinereadable
files (including this one) to any server computer, is strictly prohibited. To order Numerical Recipes books or CDROMs, visit website
http://www.nr.com or call 1-800-872-7423 (North America only), or send email to directcustserv@cambridge.org (outside North America).
#include <math.h>
#include "nrutil.h"
#define BIG 1.0e30
extern int nn;
extern float *xx,*yy,*sx,*sy,*ww,aa,offs;
float chixy(float bang)
Captive function of fitexy, returns the value of (q2 - offs) for the slope b=tan(bang).
Scaled data and offs are communicated via the global variables.
{
int j;
float ans,avex=0.0,avey=0.0,sumw=0.0,b;
b=tan(bang);
for (j=1;j<=nn;j++) {
ww[j] = SQR(b*sx[j])+SQR(sy[j]);
sumw += (ww[j] = (ww[j] < 1.0/BIG ? BIG : 1.0/ww[j]));
avex += ww[j]*xx[j];
avey += ww[j]*yy[j];
}
avex /= sumw;
avey /= sumw;
aa=avey-b*avex;
for (ans = -offs,j=1;j<=nn;j++)
ans += ww[j]*SQR(yy[j]-aa-b*xx[j]);
return ans;
}*/
